package ESPOPS::DBI;

use strict;

BEGIN {
	use vars  qw(@ISA);
	@ISA	= qw( SPOPS::DBI::MySQL SPOPS::DBI );
}

use SPOPS 0.51;
use SPOPS::DBI;
use SPOPS::DBI::MySQL;

use constant DBI_DSN  => $ESPOPS::DBI::DSN  || 'DBI:mysql:spopstest';
use constant DBI_USER => $ESPOPS::DBI::USER || 'test';
use constant DBI_PASS => $ESPOPS::DBI::PASS || '';

my $DB;

sub global_datasource_handle {
	my $attempts = 1;
	until ( $DB && $DB->ping ) {
		$DB = DBI->connect( DBI_DSN, DBI_USER, DBI_PASS, {	PrintError => 0,
															RaiseError => 1,
															AutoCommit => 1 } );
		unless ($DB) {
			$attempts++;
			warn "DBI->connect() attempt " . $attempts . " (pid = $$): $DBI::errstr\n";
			die  "DBI->connect() failed: $DBI::errstr"	if $attempts > 10;
			sleep 1										if $attempts > 5;
		};
	}
	return $DB;
}

sub drop_table {
	my $class = shift;
# This will only work for classes whose config has been processed.
# I.e. it won't work for our ClubMembers type of class which is
# only used to create a 'links_to' table
# 	my $SQL = "DROP TABLE " .  $class->CONFIG->{base_table};

# ... so we have to do it this way instead ...
	my $conf = eval '$' . $class . '::CONF';
	if ($conf) {
		my ($alias) = grep $conf->{$_}->{class} eq $class, keys %$conf;
		my $SQL = "DROP TABLE IF EXISTS " .  $conf->{$alias}->{base_table};
		my $db = $class->global_datasource_handle;
		$db->do($SQL);
	}
	
	return 1;
}

sub create_table {
	my $class = shift;
	my $SQL = eval '$' . $class . '::TABLE_DEF';
	if ($SQL) {
		my $db = $class->global_datasource_handle;
		$db->do($SQL) or die $db->errstr;
	}
}

sub recreate_table {
	my $class = shift;
	$class->drop_table;
	$class->create_table;
}

1;
__END__

=head1 NAME

ESPOPS::DBI - Enhanced SPOPS DBI.

=head1 SYNOPSIS

  Assuming MyObject is a subclass of ESPOPS::Object ...

  use espops_config.pm          # defines $ESPOPS::DBI::DSN, etc.
  use MyObject;
  
  $dbh = MyObject->global_datasource_handle;
  
  MyObject->drop_table;
  MyObject->create_table;
  MyObject->recreate_table;

=head1 DESCRIPTION

This class is a parent of ESPOPS::Object, so all ESPOPS objects inherit
the methods defined here. The methods are for returning a connection to
the database, dropping, creating and recreating (doing both) the table
corresponding to the class.

=head1 METHODS

=over 4

=item * global_datasource_handle()

Returns (and saves in a package lexical) a DBI database handle, based on
the connection parameters specified by the package variables $DSN,
$USER, $PASS. These variables are typically defined in some
espops_config.pm which is included in a 'use' before the first 'use' of
an ESPOPS object. Tries 10 times before giving up.

=item *	drop_table()

Drops the class's base_table from the database.

=item *	create_table()

Creates the class's table based on the SQL stored in the class's
$TABLE_DEF.

=item *	recreate_table()

Does a drop_table() followed by a create_table().

=back

=head1 COPYRIGHT

Copyright (c) 2001-2002 PSERC. All rights reserved.

This library is free software; you can redistribute it and/or modify
it under the same terms as Perl itself.

=head1 AUTHORS

  Ray Zimmerman, <rz10@cornell.edu>

=head1 SEE ALSO

SPOPS(3)
SPOPS::DBI(3)
SPOPS::DBI::MySQL(3)
perl(1)

=cut
