package ESPOPS::Object;

use strict;
use vars qw($VERSION @ISA $CONF $TABLE_DEF $Revision);

BEGIN {
	$Revision = sprintf "%d.%03d", q$Revision: 1.41 $ =~ /: (\d+)\.(\d+)/;
	$VERSION = '0.52';
	@ISA = qw(ESPOPS::DBI);
	$CONF = {
		ObjectAlias => {
			class			=> 'ESPOPS::Object',
			base_table		=> 'Object',
			isa				=> \@ISA,
			field			=> [ qw/ id class / ],
			id_field		=> 'id',
			as_string_order => [ qw/ id class / ],
			increment_field => 1,
			no_security		=> 1,
		},
	};
	$TABLE_DEF = <<SQL;
CREATE TABLE Object (
	id int(11) AUTO_INCREMENT PRIMARY KEY,
	class  char(255),
	created timestamp(14) NOT NULL
)
SQL
}

sub e_has_a { return { }; }

use constant ROOT_OBJ_CLASS => __PACKAGE__;

use ESPOPS::DBI;
use SPOPS::ClassFactory;
use DBI qw( :sql_types );

##-----  Public Class Methods  -----
sub new {
    my $class = shift;
    my $p     = shift;
# 	my $self = $class->SUPER::new($p);
	
	## Since SUPER::new($p) ignores keys in $p that are defined as fields
	## in the CONFIG of a parent object, we have to do the assigning of
	## these parameters manually (or fix SPOPS to handle this internally).
	my $self = $class->SUPER::new;
	
	foreach my $field ( @{$class->all_fields} ) {
	    $self->{$field} = $p->{$field}	if defined $p->{$field};
	}
	
	$self->{class} = ref($self);

	return $self;
}

sub isa_classes {
	my $self	= shift;
	my $isa		= $self->_isa_classes;
	
	return [ sort { $isa->{$a} <=> $isa->{$b} } keys %$isa ];
}

sub inherited_fields {
	my $class	= shift;
	$class = ref($class)	if ref($class);	## get class if passed an object
	
	my $fields = [];
	foreach my $c ( @{$class->isa_classes} ) {
		next if $c eq $class;
		foreach my $field ( @{$c->field_list} ) {
			push @$fields, $field	unless $field eq $c->id_field;
		}
	}

	return $fields;
}

sub all_fields { return [ @{$_[0]->field_list}, @{$_[0]->inherited_fields} ]; }

sub all_field_types {
	my $class	= shift;
	my $p		= shift;
	$class = ref($class)	if ref($class);	## get class if passed an object
	
	my $type_info = {};
	foreach my $c ( @{$class->isa_classes} ) {
		my $c_types = $c->db_discover_types( $c->base_table, $p );
		foreach my $field ( @{$c->field_list} ) {
			$type_info->{$field} = $c_types->{$field}
				unless $field eq $c->id_field && $c ne $class; ## skip parent table ids
		}
	}

	return $type_info;
}

sub get_main_alias {
	## This method can be used to find the main alias even
	## before the class's configuration has been processed.
	my $class = shift;
	my $conf = eval '$' . $class . '::CONF';
	my ($alias) = grep $conf->{$_}->{class} eq $class, keys %$conf;

	return $alias;
}

sub get_CONF 		 { return eval '$' . $_[0] . '::CONF'; }
sub get_links_to	 { return $_[0]->get_CONF->{$_[0]->get_main_alias}->{links_to}; }

sub config_and_init {
	my $class = shift;
	
	SPOPS::ClassFactory->create( $class->_build_conf )
		unless $class->_config_processed; 

	$class->class_initialize;
}

sub fetch {
    my ( $class, $id, $p ) = @_;
    $p->{DEBUG} ||= SPOPS::DBI::DEBUG_FETCH;
    $p->{DEBUG} && _wm( 2, $p->{DEBUG}, "Trying to fetch an item of $class with ID $id and params ",
                           join " // ",
                                map { $_ . ' -> ' . ( defined( $p->{$_} ) ? $p->{$_} : '' ) }
                                    keys %{ $p } );

    # No ID, no object

    return undef  unless ( defined( $id ) and $id ne '' and $id !~ /^tmp/ );

    # Security violations bubble up to caller

    my $level = $p->{security_level};
    unless ( $p->{skip_security} ) {
        $level ||= $class->check_action_security({ id       => $id,
                                                   DEBUG    => $p->{DEBUG},
                                                   required => SPOPS::Secure::SEC_LEVEL_READ });
    }

    # Do any actions the class wants before fetching -- note that if
    # any of the actions returns undef (false), we bail.

    return undef unless ( $class->pre_fetch_action( { %{ $p }, id => $id } ) );

    my $obj = undef;

    # If we were passed the data for an object, go ahead and create
    # it; if not, check to see if we can whip up a cached object

    if ( ref $p->{data} eq 'HASH' ) {
        $obj = $class->new({ %{ $p->{data} }, skip_default_values => 1 });
    }
    else {
        $obj = $class->get_cached_object({ %{ $p }, id => $id });
        $p->{skip_cache}++;         # Set so we don't re-cache it later
    }

    unless ( ref $obj eq $class ) {
##-----  REPLACE THIS ORIGINAL CODE  -----
#         my ( $raw_fields, $select_fields ) = $class->_fetch_select_fields( $p );
##-----  WITH THIS OVERRIDING CODE  -----
		## Note: this code skips the column group and alter field stuff
		my $table_name		= $class->base_table;
		my $my_id_field		= $class->id_field;
		my $raw_fields		= [];
		my $select_fields	= [];
		my $sqltables		= [];
		my $sqlwhere		= [];
		foreach my $parent_class ( @{$class->isa_classes} ) {
			my $table		= $parent_class->table_name;
			my $id_field	= $parent_class->id_field;

			push @$sqltables, $table;	## list of tables for "FROM" clause
	
			## join tables by id field (set all equal to id field of this class)
			push @$sqlwhere,	$table . '.' . $id_field . ' = ' .
								$table_name . '.' . $my_id_field
									unless($table_name eq $table);
	
			## all fields, except id of inherited tables
			foreach my $field ( keys %{$parent_class->field} ) {
				next if $parent_class ne $class && $field eq $id_field;
				push @$select_fields, $table . '.' . $field;
				push @$raw_fields, $field;
			}
		}
		push @$sqlwhere, $class->id_clause( $id, undef, $p );
##-----  END OVERRIDING CODE  -----
        $p->{DEBUG} && _wm( 1, $p->{DEBUG}, "SELECTing: ", join( "//", @{ $select_fields } ) );

        # Put all the arguments into a hash (so we can reuse them simply
        # later) and grab the record

##-----  REPLACE THIS ORIGINAL CODE  -----
#         my %args = ( from   => [ $class->table_name ],
#                      select => $select_fields,
#                      where  => $class->id_clause( $id, undef, $p ),
##-----  WITH THIS OVERRIDING CODE  -----
		my %args = (	from   => $sqltables,
						select => $select_fields,
						where  => join(' AND ', @$sqlwhere),
##-----  END OVERRIDING CODE  -----
                     db     => $p->{db},
                     return => 'single',
                     DEBUG  => $p->{DEBUG} );
        my $row = eval { $class->db_select( \%args ) };
        if ( $@ ) {
            $class->fail_fetch( \%args );
            die $@;
        }

        # If the row isn't found, return nothing; just as if an incorrect
        # (or nonexistent) ID were passed in

        return undef unless ( $row );

        # Note that we pass $p along to the ->new() method, in case
        # other information was passed in needed by it -- however, we
        # need to be careful that certain parameters used by this
        # method (e.g., the optional 'field_alter') is not the same as
        # a parameter of an object -- THAT would be fun to debug...

        $obj = $class->new({ id => $id, skip_default_values => 1, %{ $p } });
        $obj->_fetch_assign_row( $raw_fields, $row, $p );
    }
    return $obj->_fetch_post_process( $p, $level );
}

sub fetch_group {
    my ( $class, $p ) = @_;
    $p->{DEBUG} ||= SPOPS::DBI::DEBUG_FETCH;
##-----  REPLACE THIS ORIGINAL CODE  -----
#     ( $p->{raw_fields}, $p->{select} ) = $class->_construct_group_select( $p );
##-----  WITH THIS OVERRIDING CODE  -----
	## Note: this code skips the column group and alter field stuff
	my $table_name = $class->table_name;
	my $p_original = { %$p };
	my $my_id_field		= $class->id_field;
	my $raw_fields		= [];
	my $select_fields	= [];
	my $sqltables		= [];
	my $sqlwhere		= [];

	foreach my $parent_class ( @{$class->isa_classes} ) {
		my $table		= $parent_class->table_name;
		my $id_field	= $parent_class->id_field;

		push @$sqltables, $table;	## list of tables for "FROM" clause

		## join tables by id field (set all equal to id field of this class)
		push @$sqlwhere,	$table . '.' . $id_field . ' = ' .
							$table_name . '.' . $my_id_field
								unless($table_name eq $table);

		## all fields, except id of inherited tables
		foreach my $field ( keys %{$parent_class->field} ) {
			next if $parent_class ne $class && $field eq $id_field;
			push @$select_fields, $table . '.' . $field;
			push @$raw_fields, $field;
		}
	}

	## original table list and WHERE clause
	push @$sqltables, @{ $p->{from} }	if $p->{from};
	push @$sqlwhere, $p->{where}		if $p->{where};

	$p->{where}			= join(' AND ', @$sqlwhere);
	$p->{from}   		= $sqltables;
	$p->{select} 		= $select_fields;
	$p->{raw_fields}	= $raw_fields;

	## get indices into rows of class name and object id
	my ($classname_idx) = grep $raw_fields->[$_] eq 'class', (0..$#{$raw_fields});
	my ($id_field_idx)  = grep $raw_fields->[$_] eq $class->id_field, (0..$#{$raw_fields});
##-----  END OVERRIDING CODE  -----
    my $sth              = $class->_execute_multiple_record_query( $p );
    my ( $offset, $max ) = SPOPS::Utility->determine_limit( $p->{limit} );
    my @obj_list = ();

    my $row_count = 0;
ROW:
    while ( my $row = $sth->fetchrow_arrayref ) {
##-----  BEGIN ADDITIONAL CODE  -----
		my $newclass = $row->[ $classname_idx ];
		if ($newclass eq $class) {
##-----  END ADDITIONAL CODE  -----
        my $obj = $class->new({ skip_default_values => 1 });
        $obj->_fetch_assign_row( $p->{raw_fields}, $row, $p );
        next ROW unless ( $obj );

        # Check security on the row unless overridden by
        # 'skip_security'. If the security check fails that's ok, just
        # skip the row and move on

        my $sec_level = SPOPS::Secure::SEC_LEVEL_WRITE;
        unless ( $p->{skip_security} ) {
            $sec_level = eval { $obj->check_action_security({
                                          required => SPOPS::Secure::SEC_LEVEL_READ }) };
            if ( $@ ) {
                $p->{DEBUG} && _wm( 1, $p->{DEBUG}, "Security check for object in ",
                                          "fetch_group() failed, skipping." );
                next ROW;
            }
        }

        # Not to the offset yet, so go to the next row but still increment
        # the counter so we calculate limits properly

        if ( $offset and ( $row_count < $offset ) ) {
            $row_count++;
            next ROW;
        }
        last ROW if ( $max and ( $row_count >= $max ) );
        $row_count++;

        # If we've made it down to here, we're home free; just call the
        # post_fetch callback

        next ROW unless ( $obj->_fetch_post_process( $p, $sec_level ) );
        push @obj_list, $obj;
##-----  BEGIN ADDITIONAL CODE  -----
		} else {
			my $obj = $newclass->fetch( $row->[ $id_field_idx ], $p_original);
        next ROW unless ( $obj );

		## deleted security check (done by fetch)

        # Not to the offset yet, so go to the next row but still increment
        # the counter so we calculate limits properly

        if ( $offset and ( $row_count < $offset ) ) {
            $row_count++;
            next ROW;
        }
        last ROW if ( $max and ( $row_count >= $max ) );
        $row_count++;

		## deleted _post_fetch_process (done by fetch)
        push @obj_list, $obj;
		}
##-----  END ADDITIONAL CODE  -----
    }
    $sth->finish;
    return \@obj_list;
}

sub pm_fetch {
	my ( $class, $id, $p ) = @_;

	$p->{where} = $class->id_clause( $id );
	my $obj = $class->fetch_group( $p )->[0];

	return $obj;
}

sub fetch_group_by_field {
	my ( $class, $field, $vals, $p ) = @_;
	my $where = $class->base_table . ".$field" .
				' IN (' . join(',', map('?', @$vals)) . ')';
	$where .= ' AND (' . $p->{where} . ')'	if $p->{where}; 
	$p->{where} = $where;
	unshift @{$p->{value}}, @$vals;

	my $objs = $class->fetch_group( $p );

	return $objs;
}

sub fetch_group_by_ids {
	my ( $class, $ids, $p ) = @_;

	my $unordered = $class->fetch_group_by_field( $class->id_field, $ids, $p );
	
	## order by id list
	my %obj_by_id	= map { $_->id => $_ } @$unordered;
	my @ordered		= grep { $_ } map { $obj_by_id{$_} } @$ids;

	return \@ordered;
}

sub create_unless_exists {
	my ($class, $match_fields, $init_fields) = @_;
	my $obj;

	if ($match_fields && exists $match_fields->{id}) {
		$obj = $class->fetch($match_fields->{id});
	} else {
		$match_fields = {}	unless $match_fields;
		$match_fields->{ class } = $class;
		
		my ($where, $value) = (undef, []);		
		map {
			push @$value, $match_fields->{$_}	if defined $match_fields->{$_}
		} keys %$match_fields;

		$where = join(	' AND ',
						map {
							defined $match_fields->{$_} ? $_.' = ?' : $_.' is NULL'
						} keys %$match_fields
					);
	
		$obj = $class->fetch_group({	where => $where,
										value => $value		})->[0];


	}
	unless ($obj) {
		$init_fields = $match_fields unless $init_fields;
		delete $init_fields->{id} if exists $init_fields->{id};
		$obj = $class->new($init_fields)->save;
	}

	return $obj;
}

##-----  Public Object Methods  -----
sub save {
    my ( $self, $p ) = @_;
    $p->{DEBUG} ||= SPOPS::DBI::DEBUG_SAVE;
    $p->{DEBUG} && _wm( 1, $p->{DEBUG}, "Trying to save a (", ref $self, ")" );

    # We can force save() to be an INSERT by passing in a true value
    # for the is_add parameter; otherwise, we rely on the flag within
    # SPOPS::Tie to reflect whether an object has been saved or not.

    my $is_add = ( $p->{is_add} or ! $self->saved );

    # If this is an update and it hasn't changed, we don't need to do
    # anything.

    unless ( $is_add or $self->changed ) {
        $p->{DEBUG} && _wm( 1, $p->{DEBUG}, "This object exists and has not changed. Exiting." );
        return $self;
    }

    # Check security for create/update

    my $action = ( $is_add ) ? 'create' : 'update';
    my ( $level );
    unless ( $p->{skip_security} ) {
        $level = $self->check_action_security({ required => SPOPS::Secure::SEC_LEVEL_WRITE,
                                                is_add   => $is_add });
    }
    $p->{DEBUG} && _wm( 1, $p->{DEBUG}, "Security check passed ok. Continuing." );

    # Callback for objects to do something before they're saved

    return undef unless ( $self->pre_save_action({ %{ $p },
                                                   is_add => $is_add }) );

##-----  BEGIN ADDITIONAL CODE  -----
	## get list of classes which need to be saved
	## (put ROOT_OBJ_CLASS first, and this class last)
	my $no_insert = $p->{no_insert};	## this gets converted to an empty hash
										## which causes problems the 2nd time through
	my @classes = reverse @{ $self->isa_classes };
	foreach my $class (@classes) {
		bless $self, $class;
		$p->{field} = [];
		$p->{value} = [];
		$p->{no_insert} = $no_insert;	
##-----  END ADDITIONAL CODE  -----
    # Do not include these fields in the insert/update at all. Allow
    # user to override.

    my ( $not_included );
    if ( $is_add ) {
        $not_included = $self->no_insert;
        $p->{no_insert} ||= [];
        map { $not_included->{ $_ } = 1 } @{ $p->{no_insert} };
        $p->{no_insert} = $not_included;   # blech
    }
    else {
        $not_included = $self->no_update;
        $p->{no_update} ||= [];
        map { $not_included->{ $_ } = 1 } @{ $p->{no_update} };
    }

    # Do not include these fields in the insert/update if they're not defined
    # (note that this includes blank/empty)

    my $skip_undef   = $self->skip_undef;
    $p->{skip_undef} ||= [];
    map { $skip_undef->{ $_ } = 1 } @{ $p->{skip_undef} };

FIELD:
    foreach my $field ( keys %{ $self->field } ) {
        next FIELD if ( $not_included->{ $field } );
        my $value = $self->{ $field };
        next FIELD if ( ! defined $value and $skip_undef->{ $field } );
        push @{ $p->{field} }, $field;
        push @{ $p->{value} }, $value;
    }

    # Do the insert/update based on whether the object is new; don't
    # catch the die() that might be thrown -- let that percolate

    if ( $is_add ) { $self->_save_insert( $p )  }
    else           { $self->_save_update( $p )  }
##-----  BEGIN ADDITIONAL CODE  -----
	}
##-----  END ADDITIONAL CODE  -----

    # Do any actions that need to happen after you save the object

    return undef unless ( $self->post_save_action({ %{ $p },
                                                    is_add => $is_add }) );

    # Save the newly-created/updated object to the cache

    $self->set_cached_object( $p );

    # Note the action that we've just taken (opportunity for subclasses)

    unless ( $p->{skip_log} ) {
        $self->log_action( $action, scalar $self->id );
    }

    # Set flags and return the object so we can do chained method calls

    $self->has_save;
    $self->clear_change;
    return $self;
}

sub compare {
	my $self	= shift;
	my $twin	= shift;
	my $p		= shift;

	## must be objects of the same type
	return 0 unless ref($self) eq ref($twin);
	
	## and their fields must all have the same values
	my $type_info = $self->all_field_types($p);
	foreach my $field ( @{$self->all_fields} ) {
		next if $field eq $self->id_field;
		next unless defined $self->{$field} || defined $twin->{$field};
		return 0 unless defined $self->{$field} && $twin->{$field};
		if ( ref($self->{$field}) ) {
			return 0 unless $self->{$field}->compare($twin->{$field});
		} else {
			if ( $self->_is_numeric_type($type_info->{$field}) ) {
				return 0 unless $self->{$field} == $twin->{$field};
			} else {
				return 0 unless $self->{$field} eq $twin->{$field};
			}
		}
	}

	return 1;
}

sub as_string {
	my $self = shift;
	my $tab = shift || '';
	my $fields = $self->CONFIG->{as_string_order} || $self->all_fields;
	my $msg = '';
	foreach my $field (@$fields){
		$msg .= $tab.sprintf( "%-20s: %s\n", $field, defined $self->{$field} ? $self->{$field} : '');
		my $ref = ref $self->{$field};
		next unless ($ref  && $ref ne 'HASH' && $ref ne 'ARRAY');
		$msg .= $self->{$field}->as_string("\t");
	}
	return $msg;
}

##-----  Private Class Methods  -----
sub _build_conf {
	my $class = shift;
	my $conf  = shift || {};

	# get the conf variable for the class.
	my $class_conf = $class->get_CONF;
	# get the alias for the class
	my $alias = $class->get_main_alias;

	unless (exists ($conf->{$alias})) {
		$conf->{$alias} = $class_conf->{$alias};
	}
	my $class_links = $class->get_links_to || '';
	if ($class_links) {
		foreach my $key (keys %$class_links) {
			next if ($key->_config_processed || 
			         exists $conf->{$key->get_main_alias});
			$conf = $key->_build_conf($conf);
		}  
	}#end if 
	return $conf;
}

sub _config_processed {
	no strict 'refs';
	my $CONFIG_method = *{$_[0]."::CONFIG"}{CODE};
	return ref($CONFIG_method) eq 'CODE';
}

sub _is_numeric_type {
	my $self		= shift;
	my $type_info	= shift;
	
	return grep $type_info == $_, (	SQL_NUMERIC,
									SQL_DECIMAL,
									SQL_INTEGER,
									SQL_SMALLINT,
									SQL_FLOAT,
									SQL_REAL,
									SQL_DOUBLE,
# (no longer in DBI)				SQL_BIGINT,
									SQL_TINYINT	);
}

sub _isa_classes {
	my $class	= shift;
	my $href	= shift || {};
	my $depth	= shift || 1;

	$class = ref($class)	if ref($class);	## get class if passed an object
	$href->{$class} = $depth;				## stick it as a key in the hash
	
	foreach my $parent ( @{$class->CONFIG->{isa}} ) {
		next unless $parent->isa(ROOT_OBJ_CLASS);
		$href = $parent->_isa_classes($href, $depth+1)
			unless $href->{$parent} && $href->{$parent} > $depth+1;
	}
	return $href;
}

##-----  Private Object Methods  -----
sub post_fetch_action {
	my $self	= shift;
	my $orig_p	= shift || {};

	## call the overridden post_fetch_action to handle rulesets
	$self->SUPER::post_fetch_action( $orig_p );

	foreach my $field ( keys %{ $self->e_has_a } ) {
		my $h = $self->e_has_a->{$field};
		if ( $h->{fetch} && $h->{fetch}{type} eq 'auto' ) {
			if ( my $val = $self->{$field} ) {
				my %p;
				map {
					$p{$_} = $orig_p->{$_} if defined $orig_p->{$_}
				} qw( DEBUG db );
				$self->{$field} = $h->{class}->pm_fetch($val, \%p );
			}
		}
	}

	return $self;
}

sub pre_save_action {
	my $self	= shift;
	my $orig_p	= shift || {};
	
	## call the overridden pre_save_action to handle rulesets
	$self->SUPER::pre_save_action( $orig_p );

	foreach my $field ( keys %{ $self->e_has_a } ) {
		my $h = $self->e_has_a->{ $field };
		my $val = $self->{$field};
		if ( $val && ref $val ) {
			## save if indicated
			my %p;
			map {
				$p{$_} = $orig_p->{$_} if defined $orig_p->{$_}
			} qw( DEBUG db );
			$val->save( \%p )	if $h->{fetch} && !$h->{fetch}{nosave};

			## move object to a temp field during save
			$self->{'tmp_' . $field . '_'} = $val;
			$self->{$field} = $self->{$field}->id;
		}
	}

	return $self;
}

sub post_save_action {
	my $self	= shift;
	my $orig_p	= shift || {};

	## call the overridden post_save_action to handle rulesets
	$self->SUPER::post_save_action( $orig_p );;

	foreach my $field ( keys %{ $self->e_has_a } ) {
		my $h = $self->e_has_a->{$field};
		my $val = $self->{'tmp_' . $field . '_'};
		if ( $val && ref $val ) {
			$self->{$field} = $val;
			$self->{'tmp_' . $field . '_'} = undef;
		} elsif ( $h->{fetch} && $h->{fetch}{type} eq 'auto' ) {
			if ( my $val = $self->{$field} ) {
				my %p;
				map {
					$p{$_} = $orig_p->{$_} if defined $orig_p->{$_}
				} qw( DEBUG db );
				$self->{$field} = $h->{class}->pm_fetch($val, \%p );
			}
		}
	}

	return $self;
}

sub pre_remove_action {
	my $self	= shift;
	my $orig_p	= shift || {};
	my $class	= ref($self);

	## call the overridden pre_remove_action to handle rulesets
	$self->SUPER::pre_remove_action( $orig_p );

	## auto-remove specified secondary objects
	foreach my $field ( keys %{ $self->e_has_a } ) {
		my $h = $self->e_has_a->{$field};
		if ( $h->{remove} && $h->{remove}{type} eq 'auto' && $self->{$field} ) {
			my %p;
			map {
				$p{$_} = $orig_p->{$_} if defined $orig_p->{$_}
			} qw( DEBUG db );
			$self->{$field} = $h->{class}->pm_fetch($self->{$field}, \%p )
				unless ref $self->{$field};
			$self->{$field}->remove( $orig_p );
		}
	}

	## remove all corresponding rows in 'links_to' tables
	my ($table, $where);
	foreach $table ( values %{$self->CONFIG->{links_to}} ) {
		$where = $self->id_clause(undef, 'noqualify', $orig_p);
		eval { $self->db_delete( {	table => $table,
									where => $where,
									db	  => $orig_p->{db},
									DEBUG => $orig_p->{DEBUG} }) };
		if ( $@ ) {
			warn "Unable to remove links."; 
		}
	}
	
	## remove corresponding row in each parent table
	foreach my $c ( @{$class->isa_classes} ) {
		next if $c eq $class;
		$table = $c->table_name;
		$where = $c->id_clause($self->id, 'noqualify', $orig_p);
		eval { $self->db_delete( {	table => $table,
									where => $where,
									db	  => $orig_p->{db},
									DEBUG => $orig_p->{DEBUG} }) };
		if ( $@ ) {
			warn "Unable to remove row from parent table."; 
		}
	}

	return $self;
}

__PACKAGE__->config_and_init;

1;
__END__

=head1 NAME

ESPOPS::Object - Enhanced SPOPS Object.

=head1 SYNOPSIS

Assuming MyObject is a subclass of ESPOPS::Object ...

  use MyObject;
  $obj = new MyObject;
  $obj = $obj->save;
  $obj = MyObject->fetch($id);
  $obj->remove;


=head1 DESCRIPTION

This is the base class for all ESPOPS (Enhanced SPOPS) objects. Where
standard SPOPS (or more correctly, SPOPS::DBI) implements persistence
for simple class of objects whose data is stored in a single table,
ESPOPS implements persistence for a set of classes with inheritance
relationships, where the attributes are distributed across several
tables, one for each class in the inheritance hierarchy.

For example, suppose we have a Vehicle class inheriting from
ESPOPS::Object, which has the fields 'name' and 'owner'. And suppose
Vehicle has a subclass Aircraft which adds the field 'ceiling'.
ESPOPS::Object also defines the fields 'id' and 'class'. In this
example, an Aircraft object will be stored into 3 tables, 'id' and
'class' in the base_table for ESPOPS::Object, 'name' and 'owner' in the
base_table for Vehicle and 'ceiling' in the base_table for Aircraft.
Each table also has an id_field which is used to join the pieces of the
object together from the 3 tables.

Also, unlike the typical usage of SPOPS objects, where the classes are
created by SPOPS and have no corresponding .pm file, ESPOPS objects are
defined in a .pm file just like a standard Perl object, with a few
additions. Each class must define the variables @ISA, $CONF, and
$TABLE_DEF in the BEGIN block. The @ISA variable is standard Perl and
$TABLE_DEF contains an SQL statement which creates the table for the
corresponding class. The $CONF variable contains an SPOPS configuration
hash with the configuration for this class only. The BEGIN block is
followed by 'use' statements for the classes which are referenced in
@ISA and the 'has_a' and 'links_to' parts of $CONF. Finally, after all
of the use statements, it should have the line:

  __PACKAGE__->config_and_init;

By convention we put it as the last line of code in the file.

These conventions allow us to say ...

  use MyObject;

... just like we would 'use' any other Perl object.

=head1 ATTRIBUTES

=head2 Persistent Attributes

=over 4

=item id

An auto-increment integer field which stores the object id.

=item class

A string containing the name of the object's class.

=back

=head1 METHODS

=head2 Public Class Methods

=over 4

=item new

 $object = CLASS->new( $href )

Overrides the inherited new() method to allow the input hashref to
initialize inherited fields. Also puts the class name in the 'class'
field.

=item isa_classes

 $class_list = CLASS->isa_classes
 $class_list = $object->isa_classes

Returns an arrayref of all classes in this class's inheritance hierarchy
which inherit from ESPOPS::Object, including ESPOPS::Object and the
class itself. The list is ordered by proximity to current class in the
inheritance tree, with the calling class always returned as the first
element and ESPOPS::Object as the last element in the list.

=item inherited_fields

 $field_list = CLASS->inherited_fields
 $field_list = $object->inherited_fields

Returns an arrayref of all inherited non-id field names for the current
class. Does not include fields defined in the current class, only those
from parent classes.

=item all_fields

 $field_list = CLASS->all_fields
 $field_list = $object->all_fields

Returns an arrayref of all fields for the current class, including
inherited fields.

=item all_field_types

 $href = CLASS->all_field_types( \%p )
 $href = $object->all_field_types( \%p )

Returns a hashref of SQL types for each field (including inherited
fields), where the types are those defined by DBI. Takes an optional
hashref of parameters which are passed to db_discover_types().

=item get_main_alias

 $alias = CLASS->get_main_alias

Returns the name of the main alias for a class (the key to use in
$CONF). This method can be used even before the class's configuration
has been processed.

=item get_CONF

 $conf = CLASS->get_CONF

Returns the $CONF variable for the class.

=item get_links_to

 $links_to = CLASS->get_links_to

Returns the 'links_to' configuration hashref for the class.

=item config_and_init

 CLASS->config_and_init

Processes the configuration stored in $CONF for this class (via
SPOPS::ClassFactory->create) and any classes related to it via
'links_to' relationships. Calls class_initialize (inherited). This is
typically called by putting the line ...

 __PACKAGE__->config_and_init;

... as the last line of code in classes which inherit from
ESPOPS::Object.

=item fetch

 $object = CLASS->fetch( $id, \%p )

Overrides the SPOPS::DBI::fetch method and modifies the database SELECT.
Instead of selecting from a single table it joins the tables by id to
get all fields of the object, including inherited fields. This method
currently ignores any column_group or alter_field specifications.

=item fetch_group

 $obj_list = CLASS->fetch_group( \%p )

Overrides the SPOPS::DBI::fetch_group method and modifies the database
SELECT. Instead of selecting from a single table it joins the tables by
id to get all fields of the object, including inherited fields. This
also allows the where clause to include conditions on inherited fields.
The objects fetched are of the type specified by the 'class' field in
the Object table, not simply the class used to call fetch_group. This
method currently ignores any column_group or alter_field specifications.

=item pm_fetch

 $object = CLASS->pm_fetch( $id, \%p )

A polymorphic fetch. Identical to fetch, except the class used to
perform the fetch is not the one used to call this method, but rather
the class indicated for this id in the database. Same as a fetch_group
with where clause being a simple id clause.  This method currently
ignores any column_group or alter_field specifications.

=item fetch_group_by_field

 $obj_list = CLASS->fetch_group_by_field( $field, \@vals, \%p )

Given a field name and a list of values, it fetches the objects whose
specified field matches one of the values in the list. Simply creates
the appropriate WHERE clause and calls fetch_group to return the
corresponding arrayref of objects. Additional fetch_group parameters can
be passed in the optional \%p hashref. If %p contains a 'where' field it
is put at the end of the generated WHERE clause as follows:

 '<generated where> AND (<where from %p>)'

If %p contains a 'value' field, any generated values are added to the
beginning of the list. This allows extra conditions to be added so one
can do the following to get all Vehicles owned by Bob or Sally, except
Bob's boat.

 $list = Vehicle->fetch_group_by_field( 'owner',
 										[$bob->id, $sally->id],
                                      { where => 'id != ?',
                                        value = [ $bobs_boat->id ] }
                                     )

=item fetch_group_by_ids

 $obj_list = CLASS->fetch_group_by_ids( \@id_list, \%p )

Does a C<fetch_group_by_field> on the id field and then sorts the
returned objects according to the list of ids passed in.

Note that, while the objects returned are in the same order as the
specified id's, there will be a one-to-one correspondence if and only if
all corresponding objects are in the database and are not eliminated by
an optional WHERE clause passed in %p.

=item create_unless_exists

 $obj = CLASS->create_unless_exists( \%match_fields );
 $obj = CLASS->create_unless_exists( \%match_fields, \%init_fields );

Creates an object and saves it to the database, unless a matching object
already exists in the database. Takes two optional parameters, both
hashrefs. If there is an object of this class in the database whose
corresponding fields match the values specified in %match_fields, that
object is returned. Otherwise, an new object is created using the values
specified by %init_fields, and this object is saved and returned. If
%match_fields is not defined, any existing object of the given class
will be returned. If %init_fields is not defined, the values in
%match_fields will be used to initialize any newly created object.

=back

=head2 Public Object Methods

=over 4

=item save

 $object = $object->save( \%p )

Overrides the SPOPS::DBI::save method, saving the fields of the object
in the appropriate tables. Takes an optional parameter hashref as input
and returns the saved object.

=item compare

 $TorF = $object1->compare( $object2 )

Returns 1 if the two objects contain the same data, 0 otherwise.
Note: does not compare the id fields.

=item as_string

 $str = $object->as_string

Overrides as_string method of parent class. Prints all fields, including
inherited fields, and also prints contents of nested objects (though not
fields containing hashrefs or arrayrefs).

=back

=head2 Private Class Methods

=over 4

=item _build_conf

 $href = CLASS->_build_conf( $href )

Used recursively to build up a hashref containing the configuration info
($CONF variable contents) for all classes linked to this class via
'links_to' definitions (actually only those whose config has not yet
been processed). All of these classes must be passed to
SPOPS::ClassFactory->create() at the same time to properly process the
configurations.

=item _config_processed

 $TorF = CLASS->_config_processed

Returns 1 if the class's configuration has already been processed, 0
otherwise.

=item _isa_classes

 $href = CLASS->_isa_classes
 $href = CLASS->_isa_classes($href, $depth)

Recursively builds a hashref whose keys contain the names of all classes
in this class's inheritance hierarchy which inherit from ESPOPS::Object,
including the class itself. The values in the hashref indicate level in
the inheritance hierarchy, with the calling class being level 1 and
ESPOPS::Object being the maximum.

=back

=head2 Private Object Methods

=over 4

=item pre_remove_action

 $object = $object->pre_remove_action( \%p )

Called automatically when doing $object->remove(). It removes the
corresponding row in the table corresponding to each inherited class. If
present, it also removes any corresponding rows in 'links_to' tables.

=back

=head1 BUGS / TODO

=over 4

=item *

Currently refetch() and field_update() do NOT work for inherited fields.

=item *

Make sure fetch_count works correctly. Likewise for fetch_iterator. Add
column_group and alter_field functionality back into fetch and
fetch_group.

=back

=head1 CHANGE HISTORY

=over 4

=item *

9/11/02 - (RZ) Updated to sync with SPOPS-0.68. Minor changes to fetch()
and fetch_group(). Removed docs for refetch() and field_update().

=item *

8/12/02 - (RZ) Updated to sync with SPOPS-0.64. Minor changes to fetch()
and save(). Removed refetch() and field_update() since they are now in
SPOPS::DBI.

=item *

8/3/02 - (RZ) Added create_unless_exists() (implemented by Raj).

=item *

6/21/02 - (RZ) Added refetch() and field_update() methods. Currently do
NOT work for inherited fields.

=item *

6/14/02 - (RZ) Added fetch_group_by_field() method and reimplemented
fetch_group_by_ids() to use it.

=item *

6/11/02 - (RZ) Added fetch_group_by_ids() method.

=item *

5/20/02 - (RZ) Modified isa_classes() to return classes in order of
proximity to current class in the inheritance tree. The calling class is
always the first element and ESPOPS::Object is always the last element
in the list. Modfied the order of tables in join created by
fetch/fetch_group().

=item *

5/10/02 - (RZ) Added e_has_a() method and pre/post_*_action methods to
handle forward direction auto-fetch/save/remove of secondary objects.

=item *

5/9/02 - (RZ) Added pm_fetch() method.

=item *

5/8/02 - (RZ) Updated fetch() and fetch_group() and added save() based on
versions from SPOPS::DBI in SPOPS-0.61. Updated docs.

=back

=head1 COPYRIGHT

Copyright (c) 2001-2002 PSERC. All rights reserved.

and parts are

Copyright (c) 2001-2002 intes.net, inc.. All rights reserved.

This library is free software; you can redistribute it and/or modify
it under the same terms as Perl itself.

=head1 AUTHORS

  Ray Zimmerman, <rz10@cornell.edu>
  Eliot Night, <enight@sporks.com>
  Raj Chandran, <rc264@cornell.edu>

=head1 SEE ALSO

  SPOPS(3)
  SPOPS::DBI(3)
  perl(1)

=cut
